@@ -33,8 +33,8 @@ struct SeatPrivate
     /* Configuration for this seat */
     GHashTable *properties;
 
-    /* TRUE if able to switch users */
-    gboolean can_switch;
+    /* TRUE if this seat can run multiple sessions at once */
+    gboolean supports_multi_session;
 
     /* TRUE if display server can be shared for sessions */
     gboolean share_display_server;
@@ -64,11 +64,6 @@ struct SeatPrivate
     gboolean stopped;
 };
 
-/* PAM services to use */
-#define GREETER_SERVICE   "lightdm-greeter"
-#define USER_SERVICE      "lightdm"
-#define AUTOLOGIN_SERVICE "lightdm-autologin"
-
 static void seat_logger_iface_init (LoggerInterface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (Seat, seat, G_TYPE_OBJECT,
@@ -83,7 +78,7 @@ typedef struct
 static GHashTable *seat_modules = NULL;
 
 // FIXME: Make a get_display_server() that re-uses display servers if supported
-static DisplayServer *create_display_server (Seat *seat, const gchar *session_type);
+static DisplayServer *create_display_server (Seat *seat, Session *session);
 static Greeter *create_greeter_session (Seat *seat);
 static void start_session (Seat *seat, Session *session);
 
@@ -135,6 +130,13 @@ seat_get_string_property (Seat *seat, const gchar *name)
     return g_hash_table_lookup (seat->priv->properties, name);
 }
 
+gchar **
+seat_get_string_list_property (Seat *seat, const gchar *name)
+{
+    g_return_val_if_fail (seat != NULL, NULL);
+    return g_strsplit (g_hash_table_lookup (seat->priv->properties, name), ";", 0);
+}
+
 gboolean
 seat_get_boolean_property (Seat *seat, const gchar *name)
 {
@@ -150,19 +152,29 @@ seat_get_integer_property (Seat *seat, const gchar *name)
     return value ? atoi (value) : 0;
 }
 
+const gchar *
+seat_get_name (Seat *seat)
+{
+    const gchar *name;
+
+    name = seat_get_string_property (seat, "xdg-seat");
+    if (name)
+        return name;
+
+    return "seat0";
+}
+
 void
-seat_set_can_switch (Seat *seat, gboolean can_switch)
+seat_set_supports_multi_session (Seat *seat, gboolean supports_multi_session)
 {
     g_return_if_fail (seat != NULL);
-
-    seat->priv->can_switch = can_switch;
+    seat->priv->supports_multi_session = supports_multi_session;
 }
 
 void
 seat_set_share_display_server (Seat *seat, gboolean share_display_server)
 {
     g_return_if_fail (seat != NULL);
-
     seat->priv->share_display_server = share_display_server;
 }
 
@@ -170,9 +182,13 @@ gboolean
 seat_start (Seat *seat)
 {
     g_return_val_if_fail (seat != NULL, FALSE);
-  
+
+    l_debug (seat, "Starting");
+
     SEAT_GET_CLASS (seat)->setup (seat);
-    return SEAT_GET_CLASS (seat)->start (seat);
+    seat->priv->started = SEAT_GET_CLASS (seat)->start (seat);
+
+    return seat->priv->started;
 }
 
 GList *
@@ -213,6 +229,8 @@ seat_set_active_session (Seat *seat, Session *session)
             session_lock (seat->priv->active_session);
         g_object_unref (seat->priv->active_session);
     }
+
+    session_activate (session);
     seat->priv->active_session = g_object_ref (session);
 }
 
@@ -234,7 +252,7 @@ gboolean
 seat_get_can_switch (Seat *seat)
 {
     g_return_val_if_fail (seat != NULL, FALSE);
-    return seat->priv->can_switch;
+    return seat_get_boolean_property (seat, "allow-user-switching") && seat->priv->supports_multi_session;
 }
 
 gboolean
@@ -250,7 +268,7 @@ run_script (Seat *seat, DisplayServer *display_server, const gchar *script_name,
     Process *script;
     gboolean result = FALSE;
   
-    script = process_new ();
+    script = process_new (NULL, NULL);
 
     process_set_command (script, script_name);
 
@@ -330,20 +348,20 @@ check_stopped (Seat *seat)
     }
 }
 
-static gboolean
-get_start_local_sessions (Seat *seat)
-{
-    return SEAT_GET_CLASS (seat)->get_start_local_sessions (seat);
-}
-
 static void
 display_server_stopped_cb (DisplayServer *display_server, Seat *seat)
 {
+    const gchar *script;
     GList *list, *link;
     Session *active_session;
 
     l_debug (seat, "Display server stopped");
 
+    /* Run a script right after stopping the display server */
+    script = seat_get_string_property (seat, "display-stopped-script");
+    if (script)
+        run_script (seat, NULL, script, NULL);
+
     g_signal_handlers_disconnect_matched (display_server, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, seat);
     seat->priv->display_servers = g_list_remove (seat->priv->display_servers, display_server);
 
@@ -379,7 +397,7 @@ display_server_stopped_cb (DisplayServer *display_server, Seat *seat)
     }
     g_list_free_full (list, g_object_unref);
 
-    if (!seat->priv->stopping && get_start_local_sessions (seat))
+    if (!seat->priv->stopping)
     {
         /* If we were the active session, switch to a greeter */
         active_session = seat_get_active_session (seat);
@@ -423,14 +441,13 @@ switch_to_greeter_from_failed_session (Seat *seat, Session *session)
     {
         DisplayServer *display_server;
 
-        display_server = create_display_server (seat, session_get_session_type (session));
+        display_server = create_display_server (seat, session);
+        session_set_display_server (session, display_server);
         if (!display_server_start (display_server))
         {
             l_debug (seat, "Failed to start display server for greeter");
             seat_stop (seat);
         }
-
-        session_set_display_server (session, display_server);
     }
 
     start_session (seat, SESSION (greeter_session));
@@ -502,6 +519,17 @@ run_session (Seat *seat, Session *session)
         g_object_unref (seat->priv->session_to_activate);
         seat->priv->session_to_activate = NULL;
     }
+    else if (seat->priv->active_session)
+    {
+        /* Multiple sessions can theoretically be on the same VT (especially
+           if using Mir).  If a new session appears on an existing active VT,
+           logind will mark it as active, while ConsoleKit will re-mark the
+           oldest session as active.  In either case, that may not be the
+           session that we want to be active.  So let's be explicit and
+           re-activate the correct session whenever a new session starts.
+           There's no harm to do this in seats that enforce separate VTs. */
+        session_activate (seat->priv->active_session);
+    }
 }
 
 static Session *
@@ -822,7 +850,7 @@ configure_session (Session *session, SessionConfig *config, const gchar *session
 {
     const gchar *desktop_name;
 
-    session_set_session_type (session, session_config_get_session_type (config));
+    session_set_config (session, config);
     session_set_env (session, "DESKTOP_SESSION", session_name);
     session_set_env (session, "GDMSESSION", session_name);
     desktop_name = session_config_get_desktop_name (config);
@@ -875,7 +903,7 @@ create_user_session (Seat *seat, const gchar *username, gboolean autostart)
         g_object_unref (session_config);
     }
     else
-        l_debug (seat, "Can't find session '%s'", seat_get_string_property (seat, "user-session"));
+        l_debug (seat, "Can't find session '%s'", session_name);
 
     g_object_unref (user);
 
@@ -900,20 +928,23 @@ prepend_argv (gchar ***argv, const gchar *value)
 }
 
 static Session *
-create_guest_session (Seat *seat)
+create_guest_session (Seat *seat, const gchar *session_name)
 {
-    const gchar *session_name, *guest_wrapper;
+    const gchar *guest_wrapper;
     gchar *sessions_dir, **argv;
     SessionConfig *session_config;
     Session *session;
 
-    session_name = seat_get_string_property (seat, "user-session");
+    if (!session_name)
+        session_name = seat_get_string_property (seat, "guest-session");
+    if (!session_name)
+        session_name = seat_get_string_property (seat, "user-session");
     sessions_dir = config_get_string (config_get_instance (), "LightDM", "sessions-directory");
     session_config = find_session_config (seat, sessions_dir, session_name);
     g_free (sessions_dir);
     if (!session_config)
     {
-        l_debug (seat, "Can't find session '%s'", seat_get_string_property (seat, "user-session"));
+        l_debug (seat, "Can't find session '%s'", session_name);
         return NULL;
     }
 
@@ -944,7 +975,7 @@ greeter_create_session_cb (Greeter *greeter, Seat *seat)
     Session *session;
 
     session = create_session (seat, FALSE);
-    session_set_session_type (session, session_get_session_type (SESSION (greeter)));
+    session_set_config (session, session_get_config (SESSION (greeter)));
     session_set_display_server (session, session_get_display_server (SESSION (greeter)));
 
     return g_object_ref (session);
@@ -960,10 +991,10 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
     /* Get the session to use */
     if (greeter_get_guest_authenticated (greeter))
     {
-        session = create_guest_session (seat);
+        session = create_guest_session (seat, session_name);
         if (!session)
             return FALSE;
-        session_set_pam_service (session, AUTOLOGIN_SERVICE);
+        session_set_pam_service (session, seat_get_string_property (seat, "pam-autologin-service"));
     }
     else
     {
@@ -1004,7 +1035,7 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
         g_free (sessions_dir);
         if (!session_config)
         {
-            l_debug (seat, "Can't find session '%s'", seat_get_string_property (seat, "user-session"));
+            l_debug (seat, "Can't find session '%s'", session_name);
             return FALSE;
         }
 
@@ -1050,7 +1081,7 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
     /* Otherwise start a new display server for this session */
     else
     {
-        display_server = create_display_server (seat, session_get_session_type (session));
+        display_server = create_display_server (seat, session);
         session_set_display_server (session, display_server);
         if (!display_server_start (display_server))
         {
@@ -1069,6 +1100,9 @@ create_greeter_session (Seat *seat)
     SessionConfig *session_config;
     Greeter *greeter_session;
     const gchar *greeter_wrapper;
+    const gchar *autologin_username;
+    int autologin_timeout;
+    gboolean autologin_guest;
 
     l_debug (seat, "Creating greeter session");
 
@@ -1089,7 +1123,7 @@ create_greeter_session (Seat *seat)
     }
 
     greeter_session = SEAT_GET_CLASS (seat)->create_greeter_session (seat);
-    session_set_session_type (SESSION (greeter_session), session_config_get_session_type (session_config));
+    session_set_config (SESSION (greeter_session), session_config);
     seat->priv->sessions = g_list_append (seat->priv->sessions, SESSION (greeter_session));
     g_signal_connect (greeter_session, "notify::active-username", G_CALLBACK (greeter_active_username_changed_cb), seat);
     g_signal_connect (greeter_session, "authentication-complete", G_CALLBACK (session_authentication_complete_cb), seat);
@@ -1098,7 +1132,7 @@ create_greeter_session (Seat *seat)
     set_session_env (SESSION (greeter_session));
     session_set_env (SESSION (greeter_session), "XDG_SESSION_CLASS", "greeter");
 
-    session_set_pam_service (SESSION (greeter_session), GREETER_SERVICE);
+    session_set_pam_service (SESSION (greeter_session), seat_get_string_property (seat, "pam-greeter-service"));
     if (getuid () == 0)
     {
         gchar *greeter_user;      
@@ -1114,7 +1148,9 @@ create_greeter_session (Seat *seat)
     session_set_argv (SESSION (greeter_session), argv);
     g_strfreev (argv);
 
-    greeter_set_pam_services (greeter_session, USER_SERVICE, AUTOLOGIN_SERVICE);
+    greeter_set_pam_services (greeter_session,
+                              seat_get_string_property (seat, "pam-service"),
+                              seat_get_string_property (seat, "pam-autologin-service"));
     g_signal_connect (greeter_session, "create-session", G_CALLBACK (greeter_create_session_cb), seat);
     g_signal_connect (greeter_session, "start-session", G_CALLBACK (greeter_start_session_cb), seat);
 
@@ -1126,6 +1162,25 @@ create_greeter_session (Seat *seat)
     greeter_set_hint (greeter_session, "show-remote-login", seat_get_boolean_property (seat, "greeter-show-remote-login") ? "true" : "false");
     greeter_set_hint (greeter_session, "has-guest-account", seat_get_allow_guest (seat) && seat_get_boolean_property (seat, "greeter-allow-guest") ? "true" : "false");
 
+    /* Configure for automatic login */
+    autologin_username = seat_get_string_property (seat, "autologin-user");
+    if (g_strcmp0 (autologin_username, "") == 0)
+        autologin_username = NULL;
+    autologin_timeout = seat_get_integer_property (seat, "autologin-user-timeout");
+    autologin_guest = seat_get_boolean_property (seat, "autologin-guest");
+    if (autologin_timeout > 0)
+    {
+        gchar *value;
+
+        value = g_strdup_printf ("%d", autologin_timeout);
+        greeter_set_hint (greeter_session, "autologin-timeout", value);
+        g_free (value);
+        if (autologin_username)
+            greeter_set_hint (greeter_session, "autologin-user", autologin_username);
+        if (autologin_guest)
+            greeter_set_hint (greeter_session, "autologin-guest", "true");
+    }
+
     g_object_unref (session_config);
 
     return greeter_session;
@@ -1161,10 +1216,6 @@ display_server_ready_cb (DisplayServer *display_server, Seat *seat)
         return;
     }
 
-    /* Stop if don't need to run a session */
-    if (!get_start_local_sessions (seat))
-        return;
-
     emit_upstart_signal ("login-session-start");
 
     /* Start the session waiting for this display server */
@@ -1190,13 +1241,13 @@ display_server_ready_cb (DisplayServer *display_server, Seat *seat)
 }
 
 static DisplayServer *
-create_display_server (Seat *seat, const gchar *session_type)
+create_display_server (Seat *seat, Session *session)
 {
     DisplayServer *display_server;
 
-    l_debug (seat, "Creating display server of type %s", session_type);
+    l_debug (seat, "Creating display server of type %s", session_get_session_type (session));
 
-    display_server = SEAT_GET_CLASS (seat)->create_display_server (seat, session_type);
+    display_server = SEAT_GET_CLASS (seat)->create_display_server (seat, session);
     if (!display_server)
         return NULL;
 
@@ -1230,7 +1281,7 @@ seat_switch_to_greeter (Seat *seat)
 
     g_return_val_if_fail (seat != NULL, FALSE);
 
-    if (!seat->priv->can_switch)
+    if (!seat_get_can_switch (seat))
         return FALSE;
 
     /* Switch to greeter if one open (shouldn't be though) */
@@ -1250,7 +1301,7 @@ seat_switch_to_greeter (Seat *seat)
         g_object_unref (seat->priv->session_to_activate);
     seat->priv->session_to_activate = g_object_ref (greeter_session);
 
-    display_server = create_display_server (seat, session_get_session_type (SESSION (greeter_session)));
+    display_server = create_display_server (seat, SESSION (greeter_session));
     session_set_display_server (SESSION (greeter_session), display_server);
     if (!display_server_start (display_server))
         return FALSE;
@@ -1283,7 +1334,7 @@ switch_authentication_complete_cb (Session *session, Seat *seat)
             if (seat->priv->session_to_activate)
                 g_object_unref (seat->priv->session_to_activate);
             seat->priv->session_to_activate = g_object_ref (session);
-            display_server = create_display_server (seat, session_get_session_type (session));
+            display_server = create_display_server (seat, session);
             session_set_display_server (session, display_server);
             display_server_start (display_server);
         }
@@ -1304,7 +1355,7 @@ switch_authentication_complete_cb (Session *session, Seat *seat)
         g_object_unref (seat->priv->session_to_activate);
     seat->priv->session_to_activate = g_object_ref (greeter_session);
 
-    display_server = create_display_server (seat, session_get_session_type (SESSION (greeter_session)));
+    display_server = create_display_server (seat, SESSION (greeter_session));
     session_set_display_server (SESSION (greeter_session), display_server);
     display_server_start (display_server);
 }
@@ -1317,7 +1368,7 @@ seat_switch_to_user (Seat *seat, const gchar *username, const gchar *session_nam
     g_return_val_if_fail (seat != NULL, FALSE);
     g_return_val_if_fail (username != NULL, FALSE);
 
-    if (!seat->priv->can_switch)
+    if (!seat_get_can_switch (seat))
         return FALSE;
 
     /* If we're already on this session, then ignore */
@@ -1330,10 +1381,9 @@ seat_switch_to_user (Seat *seat, const gchar *username, const gchar *session_nam
     /* Attempt to authenticate them */
     session = create_user_session (seat, username, FALSE);
     g_signal_connect (session, "authentication-complete", G_CALLBACK (switch_authentication_complete_cb), seat);
-    session_set_pam_service (session, USER_SERVICE);
-    session_start (session);
+    session_set_pam_service (session, seat_get_string_property (seat, "pam-service"));
 
-    return FALSE;
+    return session_start (session);
 }
 
 static Session *
@@ -1359,7 +1409,7 @@ seat_switch_to_guest (Seat *seat, const gchar *session_name)
 
     g_return_val_if_fail (seat != NULL, FALSE);
 
-    if (!seat->priv->can_switch || !seat_get_allow_guest (seat))
+    if (!seat_get_can_switch (seat) || !seat_get_allow_guest (seat))
         return FALSE;
 
     /* Switch to session if one open */
@@ -1371,21 +1421,19 @@ seat_switch_to_guest (Seat *seat, const gchar *session_name)
         return TRUE;
     }
 
-    session = create_guest_session (seat);
+    session = create_guest_session (seat, session_name);
     if (!session)
         return FALSE;
 
-    display_server = create_display_server (seat, session_get_session_type (session));
-    if (!display_server_start (display_server))
-        return FALSE;
+    display_server = create_display_server (seat, session);
 
     if (seat->priv->session_to_activate)
         g_object_unref (seat->priv->session_to_activate);
     seat->priv->session_to_activate = g_object_ref (session);
-    session_set_pam_service (session, AUTOLOGIN_SERVICE);
+    session_set_pam_service (session, seat_get_string_property (seat, "pam-autologin-service"));
     session_set_display_server (session, display_server);
 
-    return TRUE;
+    return display_server_start (display_server);
 }
 
 gboolean
@@ -1396,7 +1444,7 @@ seat_lock (Seat *seat, const gchar *username)
 
     g_return_val_if_fail (seat != NULL, FALSE);
 
-    if (!seat->priv->can_switch)
+    if (!seat_get_can_switch (seat))
         return FALSE;
 
     l_debug (seat, "Locking");
@@ -1414,9 +1462,7 @@ seat_lock (Seat *seat, const gchar *username)
     if (!greeter_session)
         return FALSE;
 
-    display_server = create_display_server (seat, session_get_session_type (SESSION (greeter_session)));
-    if (!display_server_start (display_server))
-        return FALSE;
+    display_server = create_display_server (seat, SESSION (greeter_session));
 
     if (seat->priv->session_to_activate)
         g_object_unref (seat->priv->session_to_activate);
@@ -1426,7 +1472,7 @@ seat_lock (Seat *seat, const gchar *username)
         greeter_set_hint (greeter_session, "select-user", username);
     session_set_display_server (SESSION (greeter_session), display_server);
 
-    return TRUE;
+    return display_server_start (display_server);
 }
 
 void
@@ -1449,12 +1495,6 @@ seat_get_is_stopping (Seat *seat)
     return seat->priv->stopping;
 }
 
-static gboolean
-seat_real_get_start_local_sessions (Seat *seat)
-{
-    return TRUE;
-}
-
 static void
 seat_real_setup (Seat *seat)
 {
@@ -1469,16 +1509,6 @@ seat_real_start (Seat *seat)
     gboolean autologin_in_background;
     Session *session = NULL, *background_session = NULL;
 
-    l_debug (seat, "Starting");
-
-    /* If this display server doesn't have a session running on it, just start it */
-    if (!get_start_local_sessions (seat))
-    {
-        DisplayServer *display_server;
-        display_server = create_display_server (seat, "x"); // FIXME: Not necessarily an X seat, but not sure what to put here
-        return display_server_start (display_server);
-    }
-
     /* Get autologin settings */
     autologin_username = seat_get_string_property (seat, "autologin-user");
     if (g_strcmp0 (autologin_username, "") == 0)
@@ -1491,12 +1521,12 @@ seat_real_start (Seat *seat)
     if (autologin_timeout == 0 || autologin_in_background)
     {
         if (autologin_guest)
-            session = create_guest_session (seat);
+            session = create_guest_session (seat, NULL);
         else if (autologin_username != NULL)
             session = create_user_session (seat, autologin_username, TRUE);
 
         if (session)
-            session_set_pam_service (session, AUTOLOGIN_SERVICE);
+            session_set_pam_service (session, seat_get_string_property (seat, "pam-autologin-service"));
 
         /* Load in background if required */
         if (autologin_in_background && session)
@@ -1513,7 +1543,7 @@ seat_real_start (Seat *seat)
                 g_object_unref (seat->priv->session_to_activate);
             seat->priv->session_to_activate = g_object_ref (session);
 
-            display_server = create_display_server (seat, session_get_session_type (session));
+            display_server = create_display_server (seat, session);
             session_set_display_server (session, display_server);
             if (!display_server || !display_server_start (display_server))
             {
@@ -1557,7 +1587,7 @@ seat_real_start (Seat *seat)
                 greeter_set_hint (greeter_session, "autologin-guest", "true");
         }
 
-        display_server = create_display_server (seat, session_get_session_type (session));
+        display_server = create_display_server (seat, session);
         session_set_display_server (session, display_server);
         if (!display_server || !display_server_start (display_server))
         {
@@ -1581,13 +1611,12 @@ seat_real_start (Seat *seat)
     {
         DisplayServer *background_display_server;
 
-        background_display_server = create_display_server (seat, session_get_session_type (background_session));
+        background_display_server = create_display_server (seat, background_session);
         session_set_display_server (background_session, background_display_server);
         if (!display_server_start (background_display_server))
             l_warning (seat, "Failed to start display server for background session");
     }
 
-    seat->priv->started = TRUE;
     return TRUE;
 }
 
@@ -1702,7 +1731,6 @@ seat_class_init (SeatClass *klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-    klass->get_start_local_sessions = seat_real_get_start_local_sessions;
     klass->setup = seat_real_setup;
     klass->start = seat_real_start;
     klass->create_greeter_session = seat_real_create_greeter_session;
