@@ -1,16 +1,22 @@
+#define _GNU_SOURCE
+#define __USE_GNU
+
 #include <stdlib.h>
+#include <stdio.h>
 #include <string.h>
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/ioctl.h>
 #include <pwd.h>
 #include <unistd.h>
 #include <dirent.h>
 #include <grp.h>
 #include <security/pam_appl.h>
 #include <fcntl.h>
-#define __USE_GNU
 #include <dlfcn.h>
+#include <utmp.h>
+#include <utmpx.h>
 #ifdef __linux__
 #include <linux/vt.h>
 #endif
@@ -18,6 +24,8 @@
 #include <xcb/xcb.h>
 #include <gio/gunixsocketaddress.h>
 
+#include "status.h"
+
 #define LOGIN_PROMPT "login:"
 
 static int console_fd = -1;
@@ -30,9 +38,23 @@ static GList *group_entries = NULL;
 static int active_vt = 7;
 
 static gboolean status_connected = FALSE;
+static GKeyFile *config;
+
+static void connect_status (void)
+{
+    if (status_connected)
+        return;
+    status_connected = TRUE;
+
+    status_connect (NULL, NULL);
+
+    config = g_key_file_new ();
+    g_key_file_load_from_file (config, g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "script", NULL), G_KEY_FILE_NONE, NULL);
+}
 
 struct pam_handle
 {
+    char *id;
     char *service_name;
     char *user;
     char *authtok;
@@ -154,9 +176,6 @@ setresuid (uid_t ruid, uid_t uuid, uid_t suid)
 static gchar *
 redirect_path (const gchar *path)
 {
-    size_t offset;
-    gboolean matches;
-
     // Don't redirect if inside the running directory
     if (g_str_has_prefix (path, g_getenv ("LIGHTDM_TEST_ROOT")))
         return g_strdup (path);
@@ -175,7 +194,13 @@ redirect_path (const gchar *path)
         return g_strdup (path);
 
     if (g_str_has_prefix (path, "/tmp"))
-        return g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "tmp", path + strlen ("tmp"), NULL);
+        return g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "tmp", path + strlen ("/tmp"), NULL);
+
+    if (g_str_has_prefix (path, "/etc/xdg"))
+        return g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "etc", "xdg", path + strlen ("/etc/xdg"), NULL);
+
+    if (g_str_has_prefix (path, "/usr/share/lightdm"))
+        return g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "usr", "share", "lightdm", path + strlen ("/usr/share/lightdm"), NULL);
 
     return g_strdup (path);
 }
@@ -251,6 +276,22 @@ fopen (const char *path, const char *mode)
     return result;
 }
 
+int
+unlinkat (int dirfd, const char *pathname, int flags)
+{
+    int (*_unlinkat) (int dirfd, const char *pathname, int flags);
+    gchar *new_path = NULL;
+    int result;
+
+    _unlinkat = (int (*)(int dirfd, const char *pathname, int flags)) dlsym (RTLD_NEXT, "unlinkat");
+
+    new_path = redirect_path (pathname);
+    result = _unlinkat (dirfd, new_path, flags);
+    g_free (new_path);
+
+    return result;
+}
+
 int
 creat (const char *pathname, mode_t mode)
 {
@@ -320,16 +361,16 @@ stat (const char *path, struct stat *buf)
 }
 
 int
-stat64 (const char *path, struct stat *buf)
+stat64 (const char *path, struct stat64 *buf)
 {
-    int (*_stat64) (const char *path, struct stat *buf);
+    int (*_stat64) (const char *path, struct stat64 *buf);
     gchar *new_path = NULL;
     int ret;
 
-    _stat64 = (int (*)(const char *path, struct stat *buf)) dlsym (RTLD_NEXT, "stat64");
+    _stat64 = (int (*)(const char *path, struct stat64 *buf)) dlsym (RTLD_NEXT, "stat64");
 
     new_path = redirect_path (path);
-    ret = _stat (new_path, buf);
+    ret = _stat64 (new_path, buf);
     g_free (new_path);
 
     return ret;
@@ -352,13 +393,13 @@ __xstat (int version, const char *path, struct stat *buf)
 }
 
 int
-__xstat64 (int version, const char *path, struct stat *buf)
+__xstat64 (int version, const char *path, struct stat64 *buf)
 {
-    int (*___xstat64) (int version, const char *path, struct stat *buf);
+    int (*___xstat64) (int version, const char *path, struct stat64 *buf);
     gchar *new_path = NULL;
     int ret;
   
-    ___xstat64 = (int (*)(int version, const char *path, struct stat *buf)) dlsym (RTLD_NEXT, "__xstat64");
+    ___xstat64 = (int (*)(int version, const char *path, struct stat64 *buf)) dlsym (RTLD_NEXT, "__xstat64");
 
     new_path = redirect_path (path);
     ret = ___xstat64 (version, new_path, buf);
@@ -367,6 +408,38 @@ __xstat64 (int version, const char *path, struct stat *buf)
     return ret;
 }
 
+int
+__fxstatat(int ver, int dirfd, const char *pathname, struct stat *buf, int flags)
+{
+    int (*___fxstatat) (int ver, int dirfd, const char *pathname, struct stat *buf, int flags);
+    gchar *new_path = NULL;
+    int ret;
+  
+    ___fxstatat = (int (*)(int ver, int dirfd, const char *pathname, struct stat *buf, int flags)) dlsym (RTLD_NEXT, "__fxstatat");
+
+    new_path = redirect_path (pathname);
+    ret = ___fxstatat (ver, dirfd, new_path, buf, flags);
+    g_free (new_path);
+
+    return ret;
+}
+
+int
+__fxstatat64(int ver, int dirfd, const char *pathname, struct stat64 *buf, int flags)
+{
+    int (*___fxstatat64) (int ver, int dirfd, const char *pathname, struct stat64 *buf, int flags);
+    gchar *new_path = NULL;
+    int ret;
+  
+    ___fxstatat64 = (int (*)(int ver, int dirfd, const char *pathname, struct stat64 *buf, int flags)) dlsym (RTLD_NEXT, "__fxstatat64");
+
+    new_path = redirect_path (pathname);
+    ret = ___fxstatat64 (ver, dirfd, new_path, buf, flags);
+    g_free (new_path);
+
+    return ret;
+}
+
 DIR *
 opendir (const char *name)
 {
@@ -402,17 +475,8 @@ mkdir (const char *pathname, mode_t mode)
 int
 chown (const char *pathname, uid_t owner, gid_t group)
 {
-    int (*_chown) (const char *pathname, uid_t owner, gid_t group);
-    gchar *new_path = NULL;
-    int result;
-
-    _chown = (int (*)(const char *pathname, uid_t owner, gid_t group)) dlsym (RTLD_NEXT, "chown");
-
-    new_path = redirect_path (pathname);
-    result = _chown (new_path, owner, group);
-    g_free (new_path);
-
-    return result;
+    /* Just fake it - we're not root */
+    return 0;
 }
 
 int
@@ -432,30 +496,33 @@ chmod (const char *path, mode_t mode)
 }
 
 int
-ioctl (int d, int request, void *data)
+ioctl (int d, unsigned long request, ...)
 {
-    int (*_ioctl) (int d, int request, void *data);
+    int (*_ioctl) (int d, int request, ...);
 
-    _ioctl = (int (*)(int d, int request, void *data)) dlsym (RTLD_NEXT, "ioctl");
+    _ioctl = (int (*)(int d, int request, ...)) dlsym (RTLD_NEXT, "ioctl");
     if (d > 0 && d == console_fd)
     {
         struct vt_stat *console_state;
-        int *n;
         int vt;
+        va_list ap;
 
         switch (request)
         {
         case VT_GETSTATE:
-            console_state = data;
+            va_start (ap, request);
+            console_state = va_arg (ap, struct vt_stat *);
+            va_end (ap);
             console_state->v_active = active_vt;
             break;
         case VT_ACTIVATE:
-            vt = GPOINTER_TO_INT (data);
+            va_start (ap, request);
+            vt = va_arg (ap, int);
+            va_end (ap);
             if (vt != active_vt)
             {
                 active_vt = vt;
-                if (!status_connected)
-                    status_connected = status_connect (NULL);
+                connect_status ();
                 status_notify ("VT ACTIVATE VT=%d", active_vt);
             }
             break;
@@ -465,7 +532,15 @@ ioctl (int d, int request, void *data)
         return 0;
     }
     else
+    {
+        va_list ap;
+        void *data;
+
+        va_start (ap, request);
+        data = va_arg (ap, void *);
+        va_end (ap);
         return _ioctl (d, request, data);
+    }
 }
 
 int
@@ -495,7 +570,7 @@ free_user (gpointer data)
 }
 
 static void
-load_passwd_file ()
+load_passwd_file (void)
 {
     gchar *path, *data = NULL, **lines;
     gint i;
@@ -627,7 +702,7 @@ free_group (gpointer data)
 }
 
 static void
-load_group_file ()
+load_group_file (void)
 {
     gchar *path, *data = NULL, **lines;
     gint i;
@@ -720,6 +795,25 @@ pam_start (const char *service_name, const char *user, const struct pam_conv *co
     if (handle == NULL)
         return PAM_BUF_ERR;
 
+    if (user)
+        handle->id = g_strdup_printf ("PAM-%s", user);
+    else
+        handle->id = g_strdup ("PAM");
+
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s START", handle->id);
+        g_string_append_printf (status, " SERVICE=%s", service_name);
+        if (user)
+            g_string_append_printf (status, " USER=%s", user);
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
+
     handle->service_name = strdup (service_name);
     handle->user = user ? strdup (user) : NULL;
     handle->authtok = NULL;
@@ -733,27 +827,6 @@ pam_start (const char *service_name, const char *user, const struct pam_conv *co
     return PAM_SUCCESS;
 }
 
-static void
-send_info (pam_handle_t *pamh, const char *message)
-{
-    struct pam_message **msg;
-    struct pam_response *resp = NULL;
-
-    msg = calloc (1, sizeof (struct pam_message *));
-    msg[0] = malloc (sizeof (struct pam_message));
-    msg[0]->msg_style = PAM_TEXT_INFO;
-    msg[0]->msg = message;
-    pamh->conversation.conv (1, (const struct pam_message **) msg, &resp, pamh->conversation.appdata_ptr);
-    free (msg[0]);
-    free (msg);
-    if (resp)
-    {
-        if (resp[0].resp)
-            free (resp[0].resp);
-        free (resp);
-    }
-}
-
 int
 pam_authenticate (pam_handle_t *pamh, int flags)
 {
@@ -762,6 +835,22 @@ pam_authenticate (pam_handle_t *pamh, int flags)
 
     if (pamh == NULL)
         return PAM_SYSTEM_ERR;
+
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s AUTHENTICATE", pamh->id);
+        if (flags & PAM_SILENT)
+            g_string_append (status, " SILENT");
+        if (flags & PAM_DISALLOW_NULL_AUTHTOK)
+            g_string_append (status, " DISALLOW_NULL_AUTHTOK");
+
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
   
     if (strcmp (pamh->service_name, "test-remote") == 0)
     {
@@ -855,9 +944,6 @@ pam_authenticate (pam_handle_t *pamh, int flags)
         free (resp);
     }
 
-    if (strcmp (pamh->user, "log-pam") == 0)
-        send_info (pamh, "pam_authenticate");
-
     /* Crash on authenticate */
     if (strcmp (pamh->user, "crash-authenticate") == 0)
         kill (getpid (), SIGSEGV);
@@ -1124,15 +1210,29 @@ pam_get_item (const pam_handle_t *pamh, int item_type, const void **item)
 int
 pam_open_session (pam_handle_t *pamh, int flags)
 {
+    GVariant *result;
+    GError *error = NULL;
+
     if (pamh == NULL)
         return PAM_SYSTEM_ERR;
 
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s OPEN-SESSION", pamh->id);
+        if (flags & PAM_SILENT)
+            g_string_append (status, " SILENT");
+
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
+
     if (strcmp (pamh->user, "session-error") == 0)
         return PAM_SESSION_ERR;
 
-    if (strcmp (pamh->user, "log-pam") == 0)
-        send_info (pamh, "pam_open_session");
-
     if (strcmp (pamh->user, "make-home-dir") == 0)
     {
         struct passwd *entry;
@@ -1140,6 +1240,33 @@ pam_open_session (pam_handle_t *pamh, int flags)
         g_mkdir_with_parents (entry->pw_dir, 0755);
     }
 
+    /* Open logind session */
+    result = g_dbus_connection_call_sync (g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, NULL),
+                                          "org.freedesktop.login1",
+                                          "/org/freedesktop/login1",
+                                          "org.freedesktop.login1.Manager",
+                                          "CreateSession",
+                                          g_variant_new ("()", ""),
+                                          G_VARIANT_TYPE ("(so)"),
+                                          G_DBUS_CALL_FLAGS_NONE,
+                                          G_MAXINT,
+                                          NULL,
+                                          &error);
+    if (result)
+    {
+        gchar *e;
+        const gchar *id;
+
+        g_variant_get (result, "(&so)", &id, NULL);
+        e = g_strdup_printf ("XDG_SESSION_ID=%s", id);
+        pam_putenv (pamh, e);
+        g_free (e);
+        g_variant_unref (result);
+    }
+    else
+        g_printerr ("Failed to create logind session: %s\n", error->message);
+    g_clear_error (&error);
+
     return PAM_SUCCESS;
 }
 
@@ -1149,8 +1276,19 @@ pam_close_session (pam_handle_t *pamh, int flags)
     if (pamh == NULL)
         return PAM_SYSTEM_ERR;
 
-    if (strcmp (pamh->user, "log-pam") == 0)
-        send_info (pamh, "pam_close_session");
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s CLOSE-SESSION", pamh->id);
+        if (flags & PAM_SILENT)
+            g_string_append (status, " SILENT");
+
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
 
     return PAM_SUCCESS;
 }
@@ -1160,13 +1298,26 @@ pam_acct_mgmt (pam_handle_t *pamh, int flags)
 {
     if (pamh == NULL)
         return PAM_SYSTEM_ERR;
+
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s ACCT-MGMT", pamh->id);
+        if (flags & PAM_SILENT)
+            g_string_append (status, " SILENT");
+        if (flags & PAM_DISALLOW_NULL_AUTHTOK)
+            g_string_append (status, " DISALLOW_NULL_AUTHTOK");
+
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
   
     if (!pamh->user)
         return PAM_USER_UNKNOWN;
 
-    if (strcmp (pamh->user, "log-pam") == 0)
-        send_info (pamh, "pam_acct_mgmt");
-
     if (strcmp (pamh->user, "denied") == 0)
         return PAM_PERM_DENIED;
     if (strcmp (pamh->user, "expired") == 0)
@@ -1188,13 +1339,29 @@ pam_chauthtok (pam_handle_t *pamh, int flags)
     if (pamh == NULL)
         return PAM_SYSTEM_ERR;
 
-    if (strcmp (pamh->user, "log-pam") == 0)
-        send_info (pamh, "pam_chauthtok");
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s CHAUTHTOK", pamh->id);
+        if (flags & PAM_SILENT)
+            g_string_append (status, " SILENT");
+        if (flags & PAM_CHANGE_EXPIRED_AUTHTOK)
+            g_string_append (status, " CHANGE_EXPIRED_AUTHTOK");
+
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
 
     msg = malloc (sizeof (struct pam_message *) * 1);
     msg[0] = malloc (sizeof (struct pam_message));
     msg[0]->msg_style = PAM_PROMPT_ECHO_OFF;
-    msg[0]->msg = "Enter new password:";
+    if ((flags & PAM_CHANGE_EXPIRED_AUTHTOK) != 0)
+        msg[0]->msg = "Enter new password (expired):";
+    else
+        msg[0]->msg = "Enter new password:";
     result = pamh->conversation.conv (1, (const struct pam_message **) msg, &resp, pamh->conversation.appdata_ptr);
     free (msg[0]);
     free (msg);
@@ -1226,8 +1393,27 @@ pam_setcred (pam_handle_t *pamh, int flags)
     if (pamh == NULL)
         return PAM_SYSTEM_ERR;
 
-    if (strcmp (pamh->user, "log-pam") == 0)
-        send_info (pamh, "pam_setcred");
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s SETCRED", pamh->id);
+        if (flags & PAM_SILENT)
+            g_string_append (status, " SILENT");
+        if (flags & PAM_ESTABLISH_CRED)
+            g_string_append (status, " ESTABLISH_CRED");
+        if (flags & PAM_DELETE_CRED)
+            g_string_append (status, " DELETE_CRED");
+        if (flags & PAM_REINITIALIZE_CRED)
+            g_string_append (status, " REINITIALIZE_CRED");
+        if (flags & PAM_REFRESH_CRED)
+            g_string_append (status, " REFRESH_CRED");
+
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
 
     /* Put the test directories into the path */
     e = g_strdup_printf ("PATH=%s/tests/src/.libs:%s/tests/src:%s/tests/src:%s/src:%s", BUILDDIR, BUILDDIR, SRCDIR, BUILDDIR, pam_getenv (pamh, "PATH"));
@@ -1252,8 +1438,12 @@ pam_setcred (pam_handle_t *pamh, int flags)
         if (group)
         {
             groups_length = getgroups (0, NULL);
+            if (groups_length < 0)
+                return PAM_SYSTEM_ERR;
             groups = malloc (sizeof (gid_t) * (groups_length + 1));
             groups_length = getgroups (groups_length, groups);
+            if (groups_length < 0)
+                return PAM_SYSTEM_ERR;
             groups[groups_length] = group->gr_gid;
             groups_length++;
             setgroups (groups_length, groups);
@@ -1272,7 +1462,19 @@ pam_end (pam_handle_t *pamh, int pam_status)
 {
     if (pamh == NULL)
         return PAM_SYSTEM_ERR;
-  
+
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("");
+        g_string_append_printf (status, "%s END", pamh->id);
+        status_notify (status->str);
+        g_string_free (status, TRUE);
+    }
+
+    free (pamh->id);
     free (pamh->service_name);
     if (pamh->user)
         free (pamh->user);
@@ -1368,11 +1570,46 @@ void
 setutxent (void)
 {
 }
-  
-struct utmp *
-pututxline (struct utmp *ut)
+
+struct utmpx *
+pututxline (const struct utmpx *ut)
 {
-    return ut;
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-utmp-config", "check-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("UTMP");
+        switch (ut->ut_type)
+        {
+        case INIT_PROCESS:
+            g_string_append_printf (status, " TYPE=INIT_PROCESS");
+            break;
+        case LOGIN_PROCESS:
+            g_string_append_printf (status, " TYPE=LOGIN_PROCESS");
+            break;
+        case USER_PROCESS:
+            g_string_append_printf (status, " TYPE=USER_PROCESS");
+            break;
+        case DEAD_PROCESS:
+            g_string_append_printf (status, " TYPE=DEAD_PROCESS");
+            break;
+        default:
+            g_string_append_printf (status, " TYPE=%d", ut->ut_type);
+        }
+        if (ut->ut_line)
+            g_string_append_printf (status, " LINE=%s", ut->ut_line);
+        if (ut->ut_id)
+            g_string_append_printf (status, " ID=%s", ut->ut_id);
+        if (ut->ut_user)
+            g_string_append_printf (status, " USER=%s", ut->ut_user);
+        if (ut->ut_host)
+            g_string_append_printf (status, " HOST=%s", ut->ut_host);
+        status_notify ("%s", status->str);
+        g_string_free (status, TRUE);
+    }
+
+    return (struct utmpx *)ut;
 }
 
 void
@@ -1380,6 +1617,46 @@ endutxent (void)
 {
 }
 
+void
+updwtmp (const char *wtmp_file, const struct utmp *ut)
+{
+    connect_status ();
+    if (g_key_file_get_boolean (config, "test-utmp-config", "check-events", NULL))
+    {
+        GString *status;
+
+        status = g_string_new ("WTMP");
+        g_string_append_printf (status, " FILE=%s", wtmp_file);
+        switch (ut->ut_type)
+        {
+        case INIT_PROCESS:
+            g_string_append_printf (status, " TYPE=INIT_PROCESS");
+            break;
+        case LOGIN_PROCESS:
+            g_string_append_printf (status, " TYPE=LOGIN_PROCESS");
+            break;
+        case USER_PROCESS:
+            g_string_append_printf (status, " TYPE=USER_PROCESS");
+            break;
+        case DEAD_PROCESS:
+            g_string_append_printf (status, " TYPE=DEAD_PROCESS");
+            break;
+        default:
+            g_string_append_printf (status, " TYPE=%d", ut->ut_type);
+        }
+        if (ut->ut_line)
+            g_string_append_printf (status, " LINE=%s", ut->ut_line);
+        if (ut->ut_id)
+            g_string_append_printf (status, " ID=%s", ut->ut_id);
+        if (ut->ut_user)
+            g_string_append_printf (status, " USER=%s", ut->ut_user);
+        if (ut->ut_host)
+            g_string_append_printf (status, " HOST=%s", ut->ut_host);
+        status_notify ("%s", status->str);
+        g_string_free (status, TRUE);
+    }
+}
+
 struct xcb_connection_t
 {
     gchar *display;
@@ -1392,7 +1669,6 @@ xcb_connect_to_display_with_auth_info (const char *display, xcb_auth_info_t *aut
 {
     xcb_connection_t *c;
     gchar *socket_path;
-    GSocketAddress *address;
     GError *error = NULL;
   
     c = malloc (sizeof (xcb_connection_t));
@@ -1417,6 +1693,7 @@ xcb_connect_to_display_with_auth_info (const char *display, xcb_auth_info_t *aut
     if (c->error == 0)
     {
         gchar *d;
+        GSocketAddress *address;
 
         /* Skip the hostname, we'll assume it's localhost */
         d = g_strdup_printf (".x%s", strchr (display, ':'));
@@ -1426,6 +1703,7 @@ xcb_connect_to_display_with_auth_info (const char *display, xcb_auth_info_t *aut
         address = g_unix_socket_address_new (socket_path);
         if (!g_socket_connect (c->socket, address, NULL, &error))
             c->error = XCB_CONN_ERROR;
+        g_object_unref (address);
         if (error)
             g_printerr ("Failed to connect to X socket %s: %s\n", socket_path, error->message);
         g_free (socket_path);
@@ -1437,8 +1715,6 @@ xcb_connect_to_display_with_auth_info (const char *display, xcb_auth_info_t *aut
     {
     }
 
-    g_object_unref (address);
-
     return c;
 }
 
